#include <algorithm>
#include <map>
#include <ctime>
#include <string>
#include <sstream>
#include "constants.h"
#include "anti_dbg.h"
#include "client_state.h"

// EXT_DECLARE_GLOBALS must be used to instantiate
// the framework's assumed globals.
EXT_DECLARE_GLOBALS();

HRESULT __thiscall
EXT_CLASS::Initialize(void) {
  m_ExtMajorVersion = 2;
  m_ExtMinorVersion = 1;

  return S_OK;
}

void __thiscall
EXT_CLASS::Uninitialize(void) {

}

void
EXT_CLASS::ProcessImageHeader(void) {
  // get _IMAGE_DOS_HEADER
  ExtRemoteTyped peb = ExtNtOsInformation::GetOsPeb();
  ULONG64 peb_base = ExtNtOsInformation::GetOsPebPtr();
  ULONG64 image_base_pointer
    = peb_base + peb.GetFieldOffset("ImageBaseAddress");
  size_t dword_bytes = sizeof(DWORD);    // bytes
  size_t word_bytes = sizeof(WORD);
  ExtRemoteData image_base("image base address",
			   image_base_pointer,
			   dword_bytes);  // 32 bit. e.g., 0x400000
  ULONG64 image_base_address = image_base.GetData(dword_bytes);
  ExtRemoteTyped image_dos_header("(ntdll!_IMAGE_DOS_HEADER *)@$extin",
				  image_base_address);

  // get _IMAGE_NT_HEADERS
  ULONG64 image_nt_headers_pointer
    = image_base_address + image_dos_header.GetFieldOffset("e_lfanew");
  ExtRemoteData image_nt_headers_pointer_data("image_nt_headers address",
					      image_nt_headers_pointer,
					      dword_bytes);
  ULONG64 image_nt_headers_address
    = image_base_address + image_nt_headers_pointer_data.GetData(dword_bytes);
  ExtRemoteTyped image_nt_headers("(ntdll!_IMAGE_NT_HEADERS *)@$extin",
				  image_nt_headers_address);

  // get TimeDateStamp
  ULONG64 timedatestamp_pointer = 
    image_nt_headers_address +
    image_nt_headers.GetFieldOffset("FileHeader") +
    image_nt_headers.Field("FileHeader").GetFieldOffset("TimeDateStamp");
  ExtRemoteData time_date_stamp("TimeDataStamp",
				timedatestamp_pointer,
				dword_bytes);
  m_stat_file << "[Info] " << std::hex << timedatestamp_pointer << " "
	      << "TimeDateStamp:"
	      << std::dec << time_date_stamp.GetData(dword_bytes) << std::endl;

  // get NumberOfSections
  ULONG64 num_sections_pointer = 
    image_nt_headers_address +
    image_nt_headers.GetFieldOffset("FileHeader") +
    image_nt_headers.Field("FileHeader")
    .GetFieldOffset("NumberOfSections");
  ExtRemoteData num_sections("number of sections",
			     num_sections_pointer,
			     word_bytes);
  for (int i = 0; i < num_sections.GetData(word_bytes); ++i) {
    ULONG64 image_section_header_address
      = image_nt_headers_address + sizeof(_IMAGE_NT_HEADERS)
      + i * sizeof(_IMAGE_SECTION_HEADER);
    char section_name[IMAGE_SIZEOF_SHORT_NAME];
    m_Data->ReadVirtualUncached(image_section_header_address,
				&section_name,
				IMAGE_SIZEOF_SHORT_NAME,  // bytes
				NULL);
    std::string name(section_name);
    if (".text" == name) {
      m_Data->ReadVirtual(image_section_header_address
			  + IMAGE_SIZEOF_SHORT_NAME
			  + sizeof(DWORD),
			  &m_text_section,
			  4,
			  NULL);
      m_text_section += image_base_address;
      m_stat_file << "[Info] " << std::hex
		  << image_section_header_address + IMAGE_SIZEOF_SHORT_NAME
		  << " .text:" << m_text_section << std::endl;
    } else if (".rdata" == name) {
      m_Data->ReadVirtual(image_section_header_address
			  + IMAGE_SIZEOF_SHORT_NAME
			  + sizeof(DWORD),
			  &m_rdata_section,
			  4,
			  NULL);
      m_rdata_section += image_base_address;
      m_stat_file << "[Info] " << std::hex
		  << image_section_header_address + IMAGE_SIZEOF_SHORT_NAME
		  << " .rdata:"
		  << m_rdata_section << std::endl;
    } else if (".data" == name) {
      m_Data->ReadVirtual(image_section_header_address
			  + IMAGE_SIZEOF_SHORT_NAME
			  + sizeof(DWORD),
			  &m_data_section,
			  4,
			  NULL);
      m_data_section += image_base_address;
      m_stat_file << "[Info] " << std::hex
		  << image_section_header_address + IMAGE_SIZEOF_SHORT_NAME
		  << " .data:"
		  << m_data_section << std::endl;
    }
  }

  if (m_data_section) {  // .data section exists
    m_memory_start = m_data_section;  // for logging purpose
  } else if (m_rdata_section){  // .rdata section exists
    m_memory_start = m_rdata_section;
  } else {  // use default values, may be wrong?
    m_memory_start = 0x00403000;
  }
  
  // get NumberOfRvaAndSizes
  ULONG64 number_pointer =
    image_nt_headers_address +
    image_nt_headers.GetFieldOffset("OptionalHeader") +
    image_nt_headers.Field("OptionalHeader")
    .GetFieldOffset("NumberOfRvaAndSizes");
  ExtRemoteData number("number of Rva and Sizes", number_pointer, dword_bytes);
  if (number.GetData(dword_bytes) >= 10) {
    // TLS entry is the 10th element in DataDirectory
    // get TLS directory RVA and Size
    ULONG64 tls_rva_pointer = number_pointer + dword_bytes + 9 * 2 * dword_bytes;
    ExtRemoteData tls_rva_data("tls rva", tls_rva_pointer, dword_bytes);
    ExtRemoteData tls_size_data("tls size",
				tls_rva_pointer + dword_bytes,
				dword_bytes);
    ULONG64 tls_rva = tls_rva_data.GetData(4);
    ULONG64 tls_size = tls_size_data.GetData(4);  // should be 0x18 in general
    if (!tls_rva) {
      return;    // not TLS callbacks
    }
    
    // find all TLS callback addresses
    ULONG64 tls_callback_address_pointer
      = image_base_address + tls_rva + 3 * dword_bytes;  // the 4th entry

    if (!tls_callback_address_pointer) {
      return;
    }
    ExtRemoteData tls_callback_address_data("first tls callback address",
					    tls_callback_address_pointer,
					    dword_bytes);
    ULONG64 first_tls_callback_address
      = tls_callback_address_data.GetData(dword_bytes);  // 1st TLS callback
    if (!first_tls_callback_address) {
      return;
    }
    for (int i = 0; ; ++i) {
      ExtRemoteData tls_callback("tls callback address",
				 first_tls_callback_address + i * dword_bytes,
				 dword_bytes);
      ULONG64 tls_callback_address = tls_callback.GetData(dword_bytes);
      if (tls_callback_address > 0x40000 && tls_callback_address < 0x10000000) {
	// set a breakpoint here, and maintain the origianl code byte
	// check if this tls entry point breakpoint has been set up
	VectorBpAddressId::iterator iter_tls = m_TLS_entries.begin();
	while (m_TLS_entries.end() != iter_tls) {
	  if ((*iter_tls).Address == tls_callback_address) {
	    break;
	  } else {
	    ++iter_tls;
	  }
	}
	
	if (m_TLS_entries.end() == iter_tls) {
	  // this tls entry point must be a new one
	  // set a breakpoint here
	  ULONG bp_id
	    = m_BreakpointEx.AddBreakpointOffset(m_Control,
						 tls_callback_address,
						 DEBUG_BREAKPOINT_CODE,
						 ENUM_BP_2_STR[TLS_CALLBACK]);

	  // make a note
	  m_stat_file << "[Anti] "
		      << std::hex << tls_callback_address
		      << " Set bp at TLS callback" << std::endl;
	  Int3AddressCode addr_code;
	  addr_code.Address = tls_callback_address;
	  addr_code.Id = bp_id;
	  m_int3_address_code.push_back(addr_code);
	  
	  // update tls entries
	  BpAddressId bp_addr_id;
	  bp_addr_id.Address = tls_callback_address;
	  bp_addr_id.Id = bp_id;
	  m_TLS_entries.push_back(bp_addr_id);
	}
      } else {
	break;
      }
    }
  }  // end of all TLS callbacks
}

void
EXT_CLASS::Init(PCSTR log_path) {
  // TODO: read from a configuration file
  // set up log file
  std::string path(log_path);  // guaranteed to be present
  m_code_file.open(path + "_code.txt", std::ios::out | std::ios::trunc);
  if (m_EnableTrace) {
    m_trace_file.open(path + "_trace.txt", std::ios::out | std::ios::trunc);
  }
  m_stat_file.open(path + "_stat.txt", std::ios::out | std::ios::trunc);

  /*
  if (m_EnableTrace) {
    m_debug_file.open(path + "_debug.txt", std::ios::out | std::ios::trunc);
  }
  */

  // get register index
  m_Registers->GetIndexByName("eax", &m_reg_id_eax);
  m_Registers->GetIndexByName("ecx", &m_reg_id_ecx);
  m_Registers->GetIndexByName("edx", &m_reg_id_edx);
  m_Registers->GetIndexByName("ebx", &m_reg_id_ebx);
  m_Registers->GetIndexByName("esp", &m_reg_id_esp);
  
  m_stack_dword_total = 128;   // decimal, 15
  m_memory_dword_total = 256;  // decimal, 40

  m_text_section = 0;    // 00401000
  m_rdata_section = 0;   // 00402000
  m_data_section = 0;    // 00403000
  
  // check if there are TLS in image header
  ProcessImageHeader();
  
  // clean any debugging bits
  CleanDebuggingTrace();    // e.g., IsDebuggerPresent()
  
  // get the entry address
  ULONG64 entry_address = GetPseudoRegisterU64("$exentry");

  // set breakpoint on entry point (may point to a unpacker)
  m_BreakpointEx.AddBreakpointOffset(m_Control,
				     entry_address,
				     DEBUG_BREAKPOINT_CODE,
				     ENUM_BP_2_STR[ENTRY]);
  m_stat_file << "[Info] "
	      << std::hex << entry_address
	      << " Set bp at $exentry" << std::endl;
    
  // set up callbacks on debug event
  m_Client->SetEventCallbacks(&m_EventCallbacks);  // register to m_Client
  m_EventCallbacks.m_Control = m_Control;
  m_EventCallbacks.m_stat_file = &m_stat_file;

  // set up disassemlby options
  ULONG flags = DEBUG_ASMOPT_IGNORE_OUTPUT_WIDTH;
  m_Control3->SetAssemblyOptions(flags);

  // prepare for single stepping
  m_tf_set_by_client = 0;
  m_pushf_by_client = false;
  m_int3_by_client = false;
  m_int3_code = 0x80000003;
  m_trap_address = 0;
  m_SEH_entry_bp_id = 0;
  m_SEH_resume_address_pointer = 0;
  
  // API related
  m_clear_eax = false;
  m_CheckRemoteDebugger_result_addr = 0;  // CheckRemoteDebuggerPresent()
  m_BlockInput_counter = 1;            // BlockInput()
  m_NtQueryInformationProcess_rlt_addr = 0;
  m_NtQueryInformationProcess_0_addr = 0;   // ProcessBasicInformation
  m_ProcessDebugFlags = false;
  m_Toolhelp32_result_addr = 0;        // Toolhelp32ReadProcessMemory
  m_Toolhelp32_original_byte = 0;
  m_UnhandledExceptFilterAddress = 0;  // SetUnhandledExceptionFilter()
  m_last_error_value = 0x1E902;        // OutoutDebugString(), random value
  m_need_to_update_last_error = false;
  m_make_CreateFile_succeeded = false;
  m_make_EndUpdateResourceA_succeeded = false;
  m_update_GetLocalTime = false;
  m_randomize_eax = false;
  m_handle_NtQueryVirtualMemory = false;
  m_mem_info = 0;             // NtQueryVirtualMemory
  m_mem_info_length = 0;      // NtQueryVirtualMemory
  m_check_eax_for_ZwQueryObject = false;  // NtQueryObject
  m_object_information_buffer = 0;        // NtQueryObject
  m_copy_gs = false;
  m_gs = 0;
  
  // there is no software breakpoint at the beginning
  m_restore_software_bp_id = 0;

  // tsc start from 0, update only when necessary
  m_time_stamp_counter = 0x10023046;  // the initial value (random)
  m_need_to_update_tsc = false;
  m_QueryPerfCounter_result_address = 0;
}

EXT_COMMAND(parse,
	    "Analyze unknown binaries.",
	    "{;}{;}{;}{;}") {
  PCSTR manual_mode = GetUnnamedArgStr(0);    // 0: manual; 1: auto
  PCSTR enable_trace = GetUnnamedArgStr(1);   // 0: no trace; 1: collect trace  
  PCSTR duration = GetUnnamedArgStr(2);       // minutes of analysis duration
  PCSTR log_path = GetUnnamedArgStr(3);       // log folder path

  if (0 == strtoul(enable_trace, 0, 10)) {
    m_EnableTrace = false;
  } else {
    m_EnableTrace = true;
  }
  
  if (0 == strtoul(manual_mode, 0, 10)) {
    m_ManualMode = false;
  } else {
    m_ManualMode = true;
  }

  m_analysis_duration = strtoul(duration, 0, 10) * 60;
  Init(log_path);    // initialize log path and handles

  std::time_t start_time = std::time(0);
  
  while (true) {
    // enter an event loop, 15 second timeout
    HRESULT status = m_Control->WaitForEvent(DEBUG_WAIT_DEFAULT, 15000);
    if (status == S_FALSE) {
      m_stat_file << "[Info] Exit:Debuggee has run freely for 5 min"
		  << std::endl;
      Exit(1);
    }
    if (std::time(0) - start_time > m_analysis_duration) {
      m_stat_file << "[Info] Exit:Achieved maximum analysis time ("
		  << std::dec << m_analysis_duration / 60
		  << " min)" << std::endl;
      Exit(1);
    }

    // examine the last event which exits the event loop
    ULONG process_id, thread_id;
    char description[300];
    ULONG des_used = 0;
    status = m_Control->GetLastEventInformation(&m_LastEventType,
						&process_id,
						&thread_id,
						&m_LastEventInfo,
						sizeof(m_LastEventInfo),
						NULL,
						description,
						300,
						&des_used);
    if (m_ManualMode) {
      m_Control->OutputPrompt(DEBUG_OUTCTL_THIS_CLIENT,
			"\nEnter !a to let ApateDbg handle: ");
      ULONG length = 0;
      m_Control->Input(&(m_UserInput[0]), 64, &length);
      std::string user_command = std::string(m_UserInput);
      while (true) {
	if ("!a" == user_command) {
	  break;
	} else if ("!g" == user_command) {
	  m_ManualMode = false;
	  break;
	} else {
	  m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
			     m_UserInput,
			     DEBUG_EXECUTE_DEFAULT);
	  m_Control->OutputPrompt(DEBUG_OUTCTL_THIS_CLIENT,
				  "\nEnter !a to let ApateDbg handle: ");
	  m_Control->Input(&(m_UserInput[0]), 64, &length);
	}
      }
    }
    
    switch (m_LastEventType) {
      case DEBUG_EVENT_BREAKPOINT: {
	// hit a breakpoint
	ULONG id = m_LastEventInfo.Breakpoint.Id;
	ULONG64 bp_address = m_BreakpointEx.GetBreakpointAddress(id);
	std::string description("");
	m_BreakpointEx.GetBpDescription(id, &description);

	if (ENUM_BP_2_STR[ENTRY] == description) {
	  m_stat_file << "[Info] "
		      << std::hex << bp_address
		      << " $exentry hit" << std::endl;
	  // at the program entry point
	  // process a single instruction
	  ProcessSingleInstruction();
	  
	  // disable the entry breakpoint
	  m_BreakpointEx.RemoveBreakpoint(m_Control, id);
	  m_Control->SetExecutionStatus(DEBUG_STATUS_GO_HANDLED);
	} else if (ENUM_BP_2_STR[EXCEPT_HANDLER] == description) {
	  // update exception record if from int 3 originally
	  if (m_int3_by_client) {
	    DEBUG_VALUE esp_pointer;
	    m_Registers->GetValue(m_reg_id_esp, &esp_pointer);
	    ULONG64 except_record_ptr = 0;
	    m_Data->ReadVirtualUncached(esp_pointer.I32 + 4,
					&except_record_ptr,
					4,  // bytes
					NULL);
	    m_Data->WriteVirtual(except_record_ptr,
				 &m_int3_code,
				 4, // bytes
				 NULL);
	    m_int3_by_client = false;
	  }
	  if (m_trap_address) {
	    DEBUG_VALUE esp_pointer;
	    m_Registers->GetValue(m_reg_id_esp, &esp_pointer);
	    ULONG64 except_record_ptr = 0;
	    m_Data->ReadVirtualUncached(esp_pointer.I32 + 4,
					&except_record_ptr,
					4,  // bytes
					NULL);
	    m_Data->WriteVirtual(except_record_ptr + 0xc,
				 &m_trap_address,
				 4, // bytes
				 NULL);
	    m_trap_address = 0;
	  }

	  // disable first and then set again after the single-stepping
	  m_stat_file << "[Info] "
		      << std::hex << bp_address
		      << " Exception handler hit" << std::endl;
	  m_BreakpointEx.DisableBreakpoint(id);    // disable for single-step
	  m_SEH_entry_bp_id = id;    // apply to SEH, Vectored
	  // save the _CONTEXT.Eip address
	  ULONG64 esp_pointer = GetRegisterU64("esp");
	  ULONG32 context_pointer = 0;
	  m_Data->ReadVirtualUncached(esp_pointer + 0xc,
				      &context_pointer,
				      4,
				      NULL);
	  m_SEH_resume_address_pointer = context_pointer + 0xb8;  // point to ip
	  m_stat_file << "[Anti] "
		      << std::hex << bp_address
		      << " Pointer to return address "
		      << "in EH saved:"
		      << std::hex << m_SEH_resume_address_pointer <<std::endl;
	  ProcessSingleInstruction();
	  m_Control->SetExecutionStatus(DEBUG_STATUS_STEP_INTO);

	  /*
	  // check if the breakpoint is in m_int3_address_code
	  for (auto it = m_int3_address_code.begin();
	       it != m_int3_address_code.end();
	       ++it) {
	    if ((*it).Address == bp_address) {
	      m_int3_address_code.erase(it);
	      break;
	    }
	  }

	  // check if the breakpoint is in m_SEH_entries
	  for (auto it = m_SEH_entries.begin();
	       it != m_SEH_entries.end();
	       ++it) {
	    if ((*it).Address == bp_address) {
	      m_SEH_entries.erase(it);
	      break;
	    }
	  }

	  // remove this breakpoint completely
	  m_BreakpointEx.RemoveBreakpoint(m_Control, id);
	  */

	} else if (ENUM_BP_2_STR[EXCEPT_RESUME_ADDRESS] == description) {
	  // we hit the resume address from an exception handler
	  // just remove this breakpoint
	  m_stat_file << "[Info] "
		      << std::hex << bp_address
		      << " Return address in EH hit"
		      << std::endl;
	  // remove for single-step
	  m_BreakpointEx.RemoveBreakpoint(m_Control, id);
	  ProcessSingleInstruction();
	  m_Control->SetExecutionStatus(DEBUG_STATUS_STEP_INTO);
	} else if (ENUM_BP_2_STR[TLS_CALLBACK] == description) {
	  m_stat_file << "[Info] "
		      << std::hex << bp_address
		      << " TLS callback hit"
		      << std::endl;
	  ProcessSingleInstruction();
	  // check if the breakpoint is in m_int3_address_code
	  bp_address = m_BreakpointEx.GetBreakpointAddress(id);
	  for (auto it = m_int3_address_code.begin();
	       it != m_int3_address_code.end();
	       ++it) {
	    if ((*it).Address == bp_address) {
	      m_int3_address_code.erase(it);
	      break;
	    }
	  }

	  // check if the breakpoint is in m_TLS_entries
	  for (auto it = m_TLS_entries.begin();
	       it != m_TLS_entries.end();
	       ++it) {
	    if ((*it).Address == bp_address) {
	      m_TLS_entries.erase(it);
	      break;
	    }
	  }

	  // remove this breakpoint completely
	  m_BreakpointEx.RemoveBreakpoint(m_Control, id);
	  m_Control->SetExecutionStatus(DEBUG_STATUS_GO_HANDLED);
	} else if (ENUM_BP_2_STR[ENUM_CALLBACK] == description) {
	  m_stat_file << "[Info] " 
		      << std::hex << bp_address
		      << " EnumFunc() callback hit" << std::endl;
	  m_BreakpointEx.RemoveBreakpoint(m_Control, id);
	  ProcessSingleInstruction();
	  // check if the breakpoint is in m_int3_address_code
	  ULONG64 bp_address = m_BreakpointEx.GetBreakpointAddress(id);
	  for (auto it = m_int3_address_code.begin();
	       it != m_int3_address_code.end();
	       ++it) {
	    if ((*it).Address == bp_address) {
	      m_int3_address_code.erase(it);
	      break;
	    }
	  }
	  m_Control->SetExecutionStatus(DEBUG_STATUS_GO_HANDLED);
	} else if (ENUM_BP_2_STR[UNHANDLED_EXCEPT_CALLBACK] == description) {
	  m_stat_file << "[Info] " << std::hex << bp_address
		      << " Unhandled exception handler hit"
		      << std::endl;
	  m_BreakpointEx.RemoveBreakpoint(m_Control, id);
	  ProcessSingleInstruction();
	  m_Control->SetExecutionStatus(DEBUG_STATUS_GO_HANDLED);
	} else {
	  // from "p", "gu", ...
	  m_stat_file << "[Info] " << std::hex << bp_address
		      << " General breakpoint hit"<< std::endl;
	  ProcessSingleInstruction();
	  m_Control->SetExecutionStatus(DEBUG_STATUS_GO_HANDLED);
	}
	break;
      }
      case DEBUG_EVENT_EXCEPTION: {
	// got an exception
	// multiple exceptions possible
	DWORD exception_code
	  = m_LastEventInfo.Exception.ExceptionRecord.ExceptionCode;
	DWORD64 exception_address
	  = m_LastEventInfo.Exception.ExceptionRecord.ExceptionAddress;

	if (!m_LastEventInfo.Exception.FirstChance) {
	  // we do not handle the second chance exception
	  // the client may wait for the unhandled exception
	  m_stat_file << "[Anti] " << std::hex << exception_address
		      << " Pass exception (2nd chance) to Client. "
		      << "Code: " << std::hex << exception_code
		      << std::endl;
	  if (m_UnhandledExceptFilterAddress) {
	    // save the _CONTEXT.Eip address
	    ULONG64 esp_pointer = GetRegisterU64("esp");
	    ULONG32 context_pointer = 0;
	    m_Data->ReadVirtualUncached(esp_pointer + 0xc,
					&context_pointer,
					4,
					NULL);
	    m_SEH_resume_address_pointer = context_pointer + 0xb8;  // point to ip
	    
	    std::stringstream command;
	    command << "r eip = 0x" << std::hex
		    << m_UnhandledExceptFilterAddress << "; ";
	    m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
			       command.str().c_str(),
			       DEBUG_EXECUTE_NO_REPEAT);
	    ULONG go_status = ProcessSingleInstruction();
	    m_Control->SetExecutionStatus(go_status);
	    break;
	  } else {
	    Exit(1);
	  }
	}

	switch(exception_code) {
	  case EXCEPTION_SINGLE_STEP: {
	    ULONG go_status = ProcessSingleInstruction();
	    m_Control->SetExecutionStatus(go_status);
	    break;
	  }
	  case EXCEPTION_BREAKPOINT: {
	    // This breakpoint is raised by client intentionally
	    // Not by our debugger
	    m_stat_file << "[Anti] " << std::hex << exception_address
			<< " Pass EXCEPTION_BREAKPOINT (1st chance) "
			<< "to Client" << std::endl;

	    m_Control->SetExecutionStatus(DEBUG_STATUS_GO_NOT_HANDLED);
	    break;
	  }
	  case EXCEPTION_ACCESS_VIOLATION: {
	    m_stat_file << "[Anti] " <<	std::hex << exception_address
			<< " Pass EXCEPTION_ACCESS_VIOLATION (1st chance) "
			<< "to Client" << std::endl;
	    m_Control->SetExecutionStatus(DEBUG_STATUS_GO_NOT_HANDLED);
	    break;
	  }
	  case EXCEPTION_INVALID_HANDLE: {
	    m_stat_file << "[Anti] " << std::hex << exception_address
			<< " Consume EXCEPTION_INVALID_HANDLE (1st chance)"
			<< std::endl;
	    // Do not pass this exception to the debuggee
	    m_Control->SetExecutionStatus(DEBUG_STATUS_GO_HANDLED);
	    ProcessSingleInstruction();  // the address will be in kernel
	    break;
	  }
	  default: {
	    m_stat_file << "[Anti] " << std::hex << exception_address
			<< " Pass Unknown exception (1st chance) to Client:"
			<< std::hex << exception_code
			<< std::endl;
	    m_Control->SetExecutionStatus(DEBUG_STATUS_GO_NOT_HANDLED);
	    break;
	  }
	}
	break;
      }
      case DEBUG_EVENT_CREATE_THREAD: {
	m_stat_file << "[Info] New thread created. "
		    << "Process_id:" << std::hex << process_id << ". "
		    << "Thread_id:" << std::hex << thread_id << ". "
		    << "Description:" << description << std::endl;
	break;
      }
      case DEBUG_EVENT_EXIT_THREAD: {
	m_stat_file << "[Info] Thread has exited." << std::endl;
	break;
      }
      case DEBUG_EVENT_CREATE_PROCESS: {
	m_stat_file << "[Info] New process created. "
		    << "Process_id:" << std::hex << process_id << ". "
		    << "Thread_id:" << std::hex << thread_id << ". "
		    << "Description-" << description << std::endl;
	break;
      }
      case DEBUG_EVENT_EXIT_PROCESS: {
	m_stat_file << "[Info] Client has exited." << std::endl;
	Exit(0);    // exit windbg
      }
      case DEBUG_EVENT_LOAD_MODULE: {
	break;
      }
      case DEBUG_EVENT_UNLOAD_MODULE: {
	break;
      }
      case DEBUG_EVENT_SYSTEM_ERROR: {
	m_stat_file << "[Info] Unhandled last event type: SYSTEM_ERROR."
		    << std::endl;
	break;
      }
      case DEBUG_EVENT_SESSION_STATUS: {
	m_stat_file << "[Info] Unhandled last event type: SESSION_STATUS."
		    << std::endl;
	break;
      }
      case DEBUG_EVENT_CHANGE_DEBUGGEE_STATE: {
	m_stat_file << "[Info] Unhandled last event type: "
		    << "CHANGE_DEBUGGEE_STATE."
		    << std::endl;
	break;
      }
      case DEBUG_EVENT_CHANGE_ENGINE_STATE: {
	m_stat_file << "[Info] unhandled last event type: CHANGE_ENGINE_STATE."
		    << std::endl;
	break;
      }
      case DEBUG_EVENT_CHANGE_SYMBOL_STATE: {
	m_stat_file << "[Info] unhandled last event type: CHANGE_SYMBOL_STATE."
		    << std::endl;
	break;
      }
      default: {
	ULONG status = ProcessSingleInstruction();
	m_Control->SetExecutionStatus(status);
      }
    }
  }
}

void
EXT_CLASS::CleanDebuggingTrace(void) {
  // IsDebuggerPresent()
  // change the peb.BeingDebugged to 0
  m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
		     "eb $peb+2 0",
		     DEBUG_EXECUTE_NO_REPEAT);

  // NtGlobalFlags and heap flags (Flags and Forceflags)
  // are disabled using '-hd' switch for WinDbg
  // normal:
  // NtGlobalFlag: 0
  // Flags: 2 (HEAP_GROWABLE)
  // ForceFlags: 0
}

ULONG
EXT_CLASS::ProcessSingleInstruction(void) {
  ULONG64 ip = GetPseudoRegisterU64("$ip");    // x86

  ULONG status = CleanUpLastInstruction(ip);
  if (status) {
    return status;    // e.g., DEBUG_STATUS_GO_NOT_HANDLED
  }
  
  // disassemble current instruction
  // Out("current eip address: %016x.\n", ip);
  char* instruction = new char[300];
  ULONG disassembled_size = 0;
  ULONG64 end_offset = 0;
  m_Control->Disassemble(ip,
			 DEBUG_DISASM_EFFECTIVE_ADDRESS,
			 instruction,
			 300,
			 &disassembled_size,
			 &end_offset);
  instruction[disassembled_size - 2] = '\0';
  instruction[disassembled_size - 1] = '\0';
  // update current disassembly code
  // e.g., 00401010 push    offset image00400000+0x1032 (00401032)
  // e.e., 00401020 50                      push    eax
  //               1st                     2nd  3rd space_idx
  ULONG second_space_idx = 0;
  UpdateCode(instruction,
	     disassembled_size,
	     end_offset,
	     ip,
	     &second_space_idx);
  
  // get the instruction name
  ULONG third_space_idx = second_space_idx + 1;
  for (; third_space_idx < disassembled_size; ++third_space_idx) {
    if (' ' == instruction[third_space_idx]) {
	break;
    }
  }

  // handle for specific instructions
  // mov dword ptr fs:[0], esp
  if (status = HandleSehUsingFs(instruction,
				second_space_idx,
				third_space_idx,
				disassembled_size)) {
    return status;
  }

  // popf/popfd
  if (status = HandlePopf(ip, instruction, second_space_idx)) {
    return status;
  }

  // pushf
  if (status = HandlePushf(ip, instruction ,second_space_idx)) {
    return status;
  }

  // call
  if (status = HandleCall2(instruction,
			   second_space_idx,
			   third_space_idx,
			   disassembled_size,
			   end_offset,
			   ip)) {
    return status;
  }

  // cmp
  if (status = HandleCmp(ip,
			 instruction,
			 second_space_idx,
			 third_space_idx,
			 disassembled_size)) {
    return status;
  }

  // rdtsc
  if (status = HandleRdtsc(instruction, second_space_idx, ip)) {
    return status;
  }

  if (status = HandleInt(ip, instruction, second_space_idx, end_offset)) {
    return status;
  }

  if (status = HandleMov(ip, instruction, second_space_idx, third_space_idx)) {
    return status;
  }

  if (status = HandleRet(ip, instruction, second_space_idx)) {
    return status;
  }
  
  // other instructions: set up single stepping
  m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
		     "r tf = 1",    // set up the trap flag
		     DEBUG_EXECUTE_NO_REPEAT);
  
  return DEBUG_STATUS_GO_HANDLED;
}

void
EXT_CLASS::Exit(int code) {
  m_Client2->TerminateCurrentProcess();
  
  // Clean up any resources.
  if (m_Control != NULL) {
    m_Control->Release();
  }
  if (m_Data != NULL) {
    m_Data->Release();
  }
  if (m_Registers != NULL) {
    m_Registers->Release();
  }
  if (m_Symbols != NULL) {
    m_Symbols->Release();
  }
  if (m_Client != NULL) {
    //
    // Request a simple end to any current session.
    // This may or may not do anything but it isn't
    // harmful to call it.
    //
    m_Client->EndSession(DEBUG_END_PASSIVE);
    
    m_Client->Release();
  }

  // write to file
  std::map<ULONG64, char*>
    ordered_addr_code_map(m_addr_code_map.begin(), m_addr_code_map.end());
  for (auto it = ordered_addr_code_map.begin();
       it != ordered_addr_code_map.end();
       ++it) {
    m_code_file << std::hex << it->first << '\t'
		<< it->second
		<< std::endl;
  }
  
  if (m_code_file.is_open()) {
    m_code_file.close();
  }

  if (m_EnableTrace && m_trace_file.is_open()) {
    m_trace_file.close();
  }
  
  if (m_stat_file.is_open()) {
    m_stat_file.close();
  }

  /*
  if (m_EnableTrace && m_debug_file.is_open()) {
    m_debug_file.close();
  }
  */
  exit(code);
}

ULONG
EXT_CLASS::UpdateCode(_In_ char* instruction,
		      _In_ ULONG disassembled_size,
		      _In_ ULONG64 end_offset,
		      _In_ ULONG64 ip,
		      _Out_opt_ ULONG* second_space_idx) {
  // e.g., 00401020 50                      push    eax
  //               1st                     2nd  3rd space_idx
  ULONG64 address = 0;
  ULONG first_space_idx = 3;
  for (; first_space_idx < disassembled_size; ++first_space_idx) {
    if (' ' == instruction[first_space_idx]) {
      break;    // 1st points to the space before the code byte 50
    }
  }
  instruction[first_space_idx] = '\0';
  address = strtoul(instruction, 0, 16);

  // the length of the 1st space series is assumed to be 1
  ULONG i = first_space_idx + 1;
  for (; i < disassembled_size; ++i) {
    if (' ' == instruction[i]) {
      break;    // i points to the first space after the code byte 50
    }
  }
  instruction[i] = '\0';  // end of code byte
  for (; i < disassembled_size; ++i) {
    if (' ' != instruction[i + 1]) {
      break;    // i points to the space before the instruction name "push"
    }
  }
  if (second_space_idx) {
    *second_space_idx = i;
  }

  // add to execution trace if in user space
  if (m_EnableTrace && address < 0x4fffffff) {
    /* light trace collection
    m_trace_file << std::hex << address << '\t'
		 << &(instruction[first_space_idx + 1])
		 << std::endl;
    */
    // save current state
    ClientState* client_state = new ClientState();
    if (!client_state) {
      m_stat_file << "[Error] Cannot new memory for client_state. Exit."
		  << std::endl;
      Exit(1);
    }
    client_state->m_eax = (DWORD)GetRegisterU64("eax");
    client_state->m_ecx = (DWORD)GetRegisterU64("ecx");
    client_state->m_edx = (DWORD)GetRegisterU64("edx");
    client_state->m_ebx = (DWORD)GetRegisterU64("ebx");
    client_state->m_esp = (DWORD)GetRegisterU64("esp");
    client_state->m_ebp = (DWORD)GetRegisterU64("ebp");
    client_state->m_esi = (DWORD)GetRegisterU64("esi");
    client_state->m_edi = (DWORD)GetRegisterU64("edi");
    client_state->m_efl = (DWORD)GetRegisterU64("efl");

    client_state->m_ES = (DWORD)GetRegisterU64("es");
    client_state->m_CS = (DWORD)GetRegisterU64("cs");
    client_state->m_SS = (DWORD)GetRegisterU64("ss");
    client_state->m_DS = (DWORD)GetRegisterU64("ds");
    client_state->m_FS = (DWORD)GetRegisterU64("fs");
    client_state->m_GS = (DWORD)GetRegisterU64("gs");
		    
    // populate stack content
    for (int i = 0; i < m_stack_dword_total; ++i) {
      // esp -> esp + 16 dword range
      DWORD stack_dword = 0;
      m_Data->ReadVirtualUncached(client_state->m_esp + 4 * i,
				  &stack_dword,
				  4,  // bytes
				  NULL);
      client_state->m_stack.push_back(stack_dword);
    }
    // populate memory content
    client_state->m_memory_start_address = m_memory_start;
    for (int i = 0; i < m_memory_dword_total; ++i) {
      DWORD memory_dword = 0;
      m_Data->ReadVirtualUncached(m_memory_start + 4 * i,
				  &memory_dword,
				  4, // bytes
				  NULL);
      client_state->m_memory.push_back(memory_dword);
    }
    
    m_trace_file << std::hex << address << '\t'
		 << &(instruction[first_space_idx + 1]) << '\t'
		 << client_state->Format()
		 << std::endl;
    delete client_state;


  }

  // insert this assembly code into code pool if not yet
  AddressCodeMap::iterator iter_addr = m_addr_code_map.find(address);
  if (m_addr_code_map.end() == iter_addr) {
    if (address < 0x4fffffff) {
      std::pair<ULONG64, char*>
	addr_code_pair (address, &(instruction[i + 1]));
      m_addr_code_map.insert(addr_code_pair);
    } else {
      // do not save kernel space code
    }
  } else {
    // TODO: update the code if different
  }

  return S_OK;
}

ULONG
EXT_CLASS::CleanUpLastInstruction(_In_ ULONG64 ip) {
  // if need to restore any software bp id
  if (m_SEH_entry_bp_id) {
    m_BreakpointEx.EnableBreakpoint(m_SEH_entry_bp_id);
    m_SEH_entry_bp_id = 0;
  }
  
  // check if need to update any states set by handling the last instruction
  if (m_restore_software_bp_id) {
    m_BreakpointEx.EnableBreakpoint(m_restore_software_bp_id);
    m_restore_software_bp_id = 0;
  }

  if (0 == m_tf_set_by_client) {
    if (m_pushf_by_client) {
      // set TF in [esp] to 0
      m_pushf_by_client = false;
      DEBUG_VALUE esp_value;
      m_Registers->GetValue(m_reg_id_esp, &esp_value);
      ULONG64 esp_pointer = esp_value.I32;
      ULONG32 stack_value= 0;
      m_Data->ReadVirtualUncached(esp_pointer,
				  &stack_value,
				  4,  // bytes
				  NULL);
      ULONG32 new_value = stack_value & 0xfffffeff;
      m_Data->WriteVirtual(esp_pointer,
			   &new_value,
			   4, // bytes
			   NULL);
      // go to analyze the coming instruction
    } else {
      // go to analyze the coming instruction
    }
  } else if (1 == m_tf_set_by_client) {
    // the last instruction is popf/popfd which sets the TF
    m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
		       "r tf = 1",    // set up the trap flag
		       DEBUG_EXECUTE_NO_REPEAT);
    ++m_tf_set_by_client;
    // go to analyze the coming instruction
  } else if (2 == m_tf_set_by_client) {
    // the one prior to the last instruction is popf/popfd which sets TF
    m_tf_set_by_client = 0;
    // whatever the last instruction is, pass the exception to the client
    m_trap_address = (DWORD)ip;
    return DEBUG_STATUS_GO_NOT_HANDLED;
  }

  if (m_clear_eax) {
    SetRegisterU64("eax", 0);
    m_clear_eax = false;
  }

  // time counter update
  if (m_need_to_update_tsc) {
    m_need_to_update_tsc = false;
    //m_time_stamp_counter += 0x2;  // add so many processor ticks
    if (!m_QueryPerfCounter_result_address) {
      SetRegisterU64("eax", m_time_stamp_counter);
    } else {
      m_Data->WriteVirtual(m_QueryPerfCounter_result_address,
			   &m_time_stamp_counter,
			   4, // bytes
			   NULL);
      m_QueryPerfCounter_result_address = 0;
    }
  }

  // CheckRemoteDebuggerPresent()
  if (m_CheckRemoteDebugger_result_addr) {
    ULONG32 result = 0;
    m_Data->WriteVirtual(m_CheckRemoteDebugger_result_addr,
			 &result,
			 4,
			 NULL);
    m_CheckRemoteDebugger_result_addr = 0;
  }

  // NtQueryInformationProcess()
  if (m_NtQueryInformationProcess_rlt_addr) {
    ULONG32 result = 0;
    if (m_ProcessDebugFlags) {
      result = 1;
      m_ProcessDebugFlags = false;
    }
    m_Data->WriteVirtual(m_NtQueryInformationProcess_rlt_addr,
			 &result,
			 4,
			 NULL);
    m_NtQueryInformationProcess_rlt_addr = 0;
  }

  // overwrite parent process id with "Explorer.exe"'s.
  if (m_NtQueryInformationProcess_0_addr) {
    PCWSTR explorer = L"explorer.exe";
    ULONG flags = DEBUG_GET_PROC_ONLY_MATCH;
    ULONG process_id = 0;
    HRESULT result =
      m_Client3->GetRunningProcessSystemIdByExecutableNameWide(0,
							       explorer,
							       flags,
							       &process_id);
    if (S_OK == result) {
      m_Data->WriteVirtual(m_NtQueryInformationProcess_0_addr + 0x14,
			   &process_id,
			   4,
			   NULL);
    }
    m_NtQueryInformationProcess_0_addr = 0;
  }

  // Toolhelp32ReadProcessMemory()
  if (m_Toolhelp32_result_addr) {
    m_Data->WriteVirtual(m_Toolhelp32_result_addr,
			 &m_Toolhelp32_original_byte,
			 1,  // write only one byte
			 NULL);
    m_Toolhelp32_result_addr = 0;
    m_Toolhelp32_original_byte = 0;
  }

  // OutputDebugString()
  if (m_need_to_update_last_error) {
    m_need_to_update_last_error = false;
    m_last_error_value += 0x1f;
    std::stringstream command;
    command << "ed fs:[34h] 0x" << std::hex << m_last_error_value << ";";
    m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
		       command.str().c_str(),  // update LastErrorValue
		       DEBUG_EXECUTE_NO_REPEAT);
  }

  // CreateFileA()
  if (m_make_CreateFile_succeeded) {
    SetRegisterU64("eax", 1);
    m_make_CreateFile_succeeded = false;
  }

  // EndUpdateResourceA()
  if (m_make_EndUpdateResourceA_succeeded) {
    SetRegisterU64("eax", 1);
    m_make_EndUpdateResourceA_succeeded = false;
  }

  // GetLoalTime()
  if (m_update_GetLocalTime) {
    m_time_stamp_counter += 0x2;  // add so many processor ticks
    ULONG32 zero = 0;
    m_Data->WriteVirtual(m_lpSystemTime + 0xa,
			 &zero,
			 4, // bytes
			 NULL);  // minute and seconds
    m_Data->WriteVirtual(m_lpSystemTime + 0xe,
			 &m_time_stamp_counter,
			 2, // bytes
			 NULL);

    m_lpSystemTime = 0;
    m_update_GetLocalTime = false;
  }

  // ZwSetDebugFilterState
  if (m_randomize_eax) {
    SetRegisterU64("eax", 0xc0000021);
    m_randomize_eax = false;
  }

  // NtQueryVirtualMemory
  if (m_handle_NtQueryVirtualMemory) {
    // find the protection field from mem_info within mem_info_length
    ULONG32 i = 0;
    while (i <= m_mem_info_length - 4) {
      ULONG32 address_protect = 0;
      m_Data->ReadVirtualUncached(m_mem_info + i, &address_protect, 4, NULL);
      ULONG32 new_protect = address_protect | 0x100;  // WSLE_PAGE_SHAREABLE
      address_protect = address_protect & 0xfffff000;
      if (address_protect == m_text_section) {
	m_Data->WriteVirtualUncached(m_mem_info + i, &new_protect, 4, NULL);
	/*
	m_stat_file << "NtQueryVirtualMemory-addr:"
		    << std::hex << m_mem_info + i << "-new_val:"
		    << std::hex << new_protect << std::endl;
	*/
	break;
      }
      i += 4;    // move to the next dword
    }
    
    m_handle_NtQueryVirtualMemory = false;
    m_mem_info = 0;
    m_mem_info_length = 0;
  }

  // NtQueryObject
  if (m_check_eax_for_ZwQueryObject) {
    if (0 == GetRegisterU64("eax")) {
      PostHandleZwQueryObject();    // handle only when it succeeded
    }
    m_check_eax_for_ZwQueryObject = false;
    m_object_information_buffer = 0;
  }

  // selectors
  if (m_copy_gs) {
    SetRegisterU64("gs", m_gs);
    SetRegisterU64("ax", m_gs);
    m_copy_gs = false;
    m_gs = 0;
  }
  
  return 0;
}

ULONG
EXT_CLASS::HandleSehUsingFs(_In_ char* instruction,
			    _In_ ULONG second_space_idx,
			    _In_ ULONG third_space_idx,
			    _In_ ULONG disassembled_size) {
  // mov    dword ptr fs:[0],esp
  for (ULONG i = third_space_idx + 1; i < disassembled_size; ++i) {
    if (' ' != instruction[i]) {
      if ('f' == instruction[i + 10] &&
	  's' == instruction[i + 11] &&
	  '[' == instruction[i + 13] &&
	  ('e' == instruction[i + 17] ||'e' == instruction[i + 19]) &&
	  'm' == instruction[second_space_idx + 1] &&
	  'o' == instruction[second_space_idx + 2]) {
	// mov     dword ptr fs:[0],esp
	// mov     dword ptr fs:[eax],esp    ; eax = 0
	// mov     dword ptr fs:[ebx],esp    ; ebx = 0
	// get the updated value
	DEBUG_VALUE except_handler_addr;
	if (('e' == instruction[i + 17] && 's' == instruction[i + 18])
	    || ('e' == instruction[i + 19] && 's' == instruction[i + 20])){
	  m_Registers->GetValue(m_reg_id_esp, &except_handler_addr);
	} else if (('e' == instruction[i + 17] && 'c' == instruction[i + 18])
		   || ('e' == instruction[i + 19]
		       && 'c' == instruction[i + 20])) {
	  m_Registers->GetValue(m_reg_id_ecx, &except_handler_addr);  // ecx
	} else if (('e' == instruction[i + 17] && 'd' == instruction[i + 18])
		   || ('e' == instruction[i + 19]
		       && 'd' == instruction[i + 20])) {
	  m_Registers->GetValue(m_reg_id_edx, &except_handler_addr);  // edx
	} else if (('e' == instruction[i + 17] && 'b' == instruction[i + 18])
		   || ('e' == instruction[i + 19]
		       && 'b' == instruction[i + 20])) {
	  m_Registers->GetValue(m_reg_id_ebx, &except_handler_addr);  // ebx
	} else if (('e' == instruction[i + 17] && 'a' == instruction[i + 18])
		   || ('e' == instruction[i + 19]
		       && 'a' == instruction[i + 20])) {
	  m_Registers->GetValue(m_reg_id_eax, &except_handler_addr);  // ebx
	} else {
	  // TODO: other registers?
	  m_stat_file << "[Todo] Unknown EH address expression: "
		      << &(instruction[second_space_idx])
		      << std::endl;
	  Exit(1);
	}

	// check if this exception handler has been set up
	VectorBpAddressId::iterator iter_seh = m_SEH_entries.begin();
	while (m_SEH_entries.end() != iter_seh) {
	  if ((*iter_seh).Address == except_handler_addr.I64) {
	    break;
	  } else {
	    ++iter_seh;
	  }
	}
	
	if (m_SEH_entries.end() == iter_seh) {
	  // this handler entry point must be a new one
	  // get the exception handler address
	  ULONG64 handler_pointer = except_handler_addr.I64 + 4;    // 32 bit
	  ULONG64 handler_entry = 0;
	  m_Data->ReadVirtualUncached(handler_pointer,
				      &handler_entry,
				      4,  // bytes
				      NULL);

	  // set a breakpoint here
	  ULONG bp_id
	    = m_BreakpointEx.AddBreakpointOffset(m_Control,
						 handler_entry,
						 DEBUG_BREAKPOINT_CODE,
						 ENUM_BP_2_STR[EXCEPT_HANDLER]);
	  m_stat_file << "[Anti] " << std::hex << handler_entry
		      << " Set a bp at SEH entry" << std::endl;
	  // make a note
	  Int3AddressCode addr_code;
	  addr_code.Address = handler_entry;
	  addr_code.Id = bp_id;
	  m_int3_address_code.push_back(addr_code);
	  
	  // update except handlers
	  BpAddressId bp_addr_id;
	  bp_addr_id.Address = except_handler_addr.I64;
	  bp_addr_id.Id = bp_id;
	  m_SEH_entries.push_back(bp_addr_id);
	} else {
	  /*
	  // handlers after this addresses are deleted
	  for (VectorBpAddressId::iterator iter = iter_seh + 1;
	       iter != m_SEH_entries.end();
	       ++iter) {
	    m_BreakpointEx.RemoveBreakpoint(m_Control, (*iter).Id);
	  }
	  m_SEH_entries.erase(iter_seh + 1, m_SEH_entries.end());
	  */
	}
	// set up single stepping
	m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
			   "r tf = 1",    // set up the trap flag
			   DEBUG_EXECUTE_NO_REPEAT);
	return DEBUG_STATUS_GO_HANDLED;
      }
      break;
    }
  }

  return 0;
}

ULONG
EXT_CLASS::HandlePopf(_In_ ULONG64 ip,
		      _In_ char* instruction,
		      _In_ ULONG second_space_idx) {
  // popf/popfd
  if (instruction[second_space_idx + 1] == 'p' &&
      instruction[second_space_idx + 2] == 'o' &&
      instruction[second_space_idx + 3] == 'p' &&
      instruction[second_space_idx + 4] == 'f') {
    // check if TF will be set by the client
    DEBUG_VALUE esp_value;
    m_Registers->GetValue(m_reg_id_esp, &esp_value);
    ULONG64 esp_pointer = esp_value.I32;
    ULONG32 stack_value= 0;
    m_Data->ReadVirtualUncached(esp_pointer,
				&stack_value,
				4,  // bytes
				NULL);
    if (stack_value & 0x00000100) {
      // trap flag will be set after executing this instruction
      m_tf_set_by_client = 1;
      m_stat_file << "[Anti] " << std::hex << ip
		  << " Client is setting Trap Flag" << std::endl;
    } 
    // set up single stepping
    m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
		       "r tf = 1",    // set up the trap flag
		       DEBUG_EXECUTE_NO_REPEAT);
    
    return DEBUG_STATUS_GO_HANDLED;
  }

  return 0;
}

ULONG
EXT_CLASS::HandlePushf(_In_ ULONG64 ip,
		       _In_ char* instruction,
		       _In_ ULONG second_space_idx) {
  // pushf/pushfd
  if ('p' == instruction[second_space_idx + 1] &&
      'u' == instruction[second_space_idx + 2] &&
      's' == instruction[second_space_idx + 3] &&
      'h' == instruction[second_space_idx + 4] &&
      'f' == instruction[second_space_idx + 5]) {
    m_stat_file << "[Anti] " << std::hex << ip
		<< " Client is reading trap flag" << std::endl;
    m_pushf_by_client = true;
    m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
		       "r tf = 1",    // set up the trap flag
		       DEBUG_EXECUTE_NO_REPEAT);
    return DEBUG_STATUS_GO_HANDLED;
  }

  return 0;
}

ULONG
EXT_CLASS::HandleCall2(_In_ char* instruction,
		       _In_ ULONG second_space_idx,
		       _In_ ULONG third_space_idx,
		       _In_ ULONG disassembled_size,
		       _In_ ULONG64 end_offset,
		       _In_ ULONG64 ip) {
  if (instruction[second_space_idx + 1] == 'c' &&
      instruction[second_space_idx + 2] == 'a' &&
      instruction[second_space_idx + 3] == 'l' &&
      instruction[second_space_idx + 4] == 'l') {
  } else {
    return 0;    // not a call instruction
  }

  // call instruction
  // check the destination address (from the right to the left)
  unsigned int i = disassembled_size - 1;
  BOOLEAN bang_found = false;
  std::string api_name;
  for (; i >= third_space_idx + 1; --i) {
    if ('!' == instruction[i]) {
      bang_found = true;
      break;    // i points to '!' now
    }
  }

  if (bang_found) {
    // get the API name
    unsigned int j = i + 1;
    while (' ' != instruction[j]) {
      ++j;
    }
    instruction[j] = '\0';
    api_name = std::string(&(instruction[i + 1]));
    instruction[j] = ' ';
  } else {
    // get the call destination
    int j = disassembled_size - 1;
    for(; j > 0; --j) {
      if (('(' == instruction[j - 1]) || ('=' == instruction[j - 1])) {
	// call    image00400000+0x100c (0040100c)
	break;
      }
    }
    BYTE original_byte = instruction[j + 8];
    instruction[j + 8] = '\0';
    ULONG64 call_dst = strtoul(&(instruction[j]), 0, 16);
    instruction[j + 8] = original_byte;

    // check the instruction at the call destination
    // e.g., jmp dword ptr [image00400000+0x200c (0040200c)]
    // ds:0023:0040200c={user32!FindWindowA (756c8ff3)}
    char* dst_instruction = new char[300];
    ULONG dst_disassembled_size = 0;
    ULONG64 dst_end_offset = 0;
    m_Control->Disassemble(call_dst,
			   DEBUG_DISASM_EFFECTIVE_ADDRESS,
			   dst_instruction,
			   300,
			   &dst_disassembled_size,
			   &dst_end_offset);
    // check if the dst_instruction is a jmp
    dst_instruction[dst_disassembled_size - 2] = '\0';
    dst_instruction[dst_disassembled_size - 1] = '\0';
    UINT32 k = 0;
    for (k = 2; k < dst_disassembled_size; ++k) {
      if (' ' == dst_instruction[k]) {
	++k;
	break;  // k points the space before code byte
      }
    }
    for (; k < dst_disassembled_size; ++k) {
      if (' ' == dst_instruction[k]) {
	break;  // k points to the space after the code byte
      }
    }
    bool is_jmp = false;
    for (; k < dst_disassembled_size; ++k) {
      if (' ' != dst_instruction[k]) {
	if ('j' == dst_instruction[k] &&
	    'm' == dst_instruction[k + 1] &&
	    'p' == dst_instruction[k + 2] &&
	    ' ' == dst_instruction[k + 3]) {
	  is_jmp = true;
	}
	break;
      }
    } // k points to the first character of the instruction name now

    if (!is_jmp) {
      // step into the call
      m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
			 "r tf = 1",    // set up the trap flag
			 DEBUG_EXECUTE_NO_REPEAT);
      return DEBUG_STATUS_STEP_INTO;
    }

    // the call dst is a jmp
    // check the jmp destination
    k = dst_disassembled_size;
    for (; k > 1; --k) {    // from right to left
      if ('!' == dst_instruction[k]) {
	bang_found = true;
	break;
      }
    }
    if (bang_found) {
      // get the API name
      unsigned int j = k + 1;
      while (' ' != dst_instruction[j]) {
	++j;
      }
      dst_instruction[j] = '\0';
      api_name = std::string(&(dst_instruction[k + 1]));
      dst_instruction[j] = ' ';

      // update the jmp instruction in code base
      UpdateCode(dst_instruction,
		 dst_disassembled_size,
		 dst_end_offset,
		 call_dst,
		 NULL);
      // not return yet.
    } else {
      // step into the call
      m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
			 "r tf = 1",    // set up the trap flag
			 DEBUG_EXECUTE_NO_REPEAT);
      return DEBUG_STATUS_STEP_INTO;
    }
  }

  m_stat_file << "[API] " << std::hex << ip << " "
	      << api_name << std::endl;

  // call dword ptr [test!_imp__AddVectoredExceptionHandler (00419004)]
  // ds:0023:00419004={ntdll!RtlAddVectoredExceptionHandler (7742f784)}
  if ("RtlAddVectoredExceptionHandler" == api_name) {
    HandleAddVectoredExceptionHandler(ip);
  } else if ("GetTickCountStub" == api_name) {
    // call    dword ptr [image01350000+0x1c004 (0136c004)]
    // ds:0023:0136c004={kernel32!GetTickCountStub (7708c340)}
    HandleGetTickCount(ip);
  } else if ("timeGetTime" == api_name) {
    // call dword ptr [test!_imp__timeGetTime (00420234)]
    // ds:0023:00420234={WINMM!timeGetTime (735126e0)}
    HandletimeGetTime(ip);
  } else if ("QueryPerformanceCounterStub" == api_name) {
    // call dword ptr [test!_imp__QueryPerformanceCounter (00420000)]
    // ds:0023:00420000={kernel32!QueryPerformanceCounterStub (7673c4d2)}
    HandleQueryPerformanceCounter(ip);
  } else if ("CheckRemoteDebuggerPresent" == api_name) {
    // call dword ptr [image00e20000+0x1c000 (00e3c000)]
    // ds:0023:00e3c000={kernel32!CheckRemoteDebuggerPresent (757b510b)}
    HandleCheckRemoteDebuggerPresent(ip);
  } else if (("FindWindowW" == api_name) || ("FindWindowA" == api_name)) {
    // call dword ptr [test!_imp__FindWindowW (00c430f8)]
    // ds:0023:00c430f8={USER32!FindWindowW (773eae0d)}
    HandleFindWindow(ip);
  } else if ("NtUserBlockInput" == api_name) {
    // call dword ptr [test!_imp__BlockInput (0042123c)]
    // ds:0023:0042123c={USER32!NtUserBlockInput (76bb6a99)}
    // skip this call by setting esp, eip, and resultant eax
    HandleBlockInput(ip, end_offset);
    ProcessSingleInstruction();
  } else if ("NtQueryInformationProcess" == api_name) {
    // call dword ptr [ebp-18h]
    // ss:0023:0012ff18={ntdll!NtQueryInformationProcess (777d6088)}
    HandleNtQueryInformationProcess(ip);
  } else if ("ZwSetInformationThread" == api_name ||
	     "NtSetInformationThread" == api_name) {
    // call dword ptr [ebp-14h]
    // ss:0023:0012ff1c={ntdll!ZwSetInformationThread (777d66d8)}
    if (0 == HandleSetInformationThread(ip, end_offset)) {
      ProcessSingleInstruction();  // ignore this call
    }  // step over otherwise
  } else if ("SwitchDesktop" == api_name) {
    // call dword ptr [test!_imp__SwitchDesktop (00420234)]
    // ds:0023:00420234={USER32!SwitchDesktop (7590476b)}
    if (0 == HandleSwitchDesktop(ip, end_offset)) {  // fake execution
      ProcessSingleInstruction();  // do not execute this call
    }  // step over otherwise
  } else if ("EnumDateFormatsWStub" == api_name
	     || "EnumDateFormatsA" == api_name
	     || "EnumDateFormatsAStub" == api_name
	     || "EnumDateFormatsExA" == api_name
	     || "EnumSystemCodePagesA" == api_name
	     || "EnumSystemLanguageGroupsA" == api_name
	     || "EnumSystemLocalesAStub" == api_name
	     || "EnumTimeFormatsA" == api_name) {
    // call dword ptr [test!_imp__EnumDateFormatsW (00420000)]
    // ds:0023:00420000={kernel32!EnumDateFormatsWStub (775d7b02)}
    // or EnumTimeFormats
    HandleEnumDateFormats(ip, 0);  // callback is the 0th parameter
  } else if ("EnumSystemGeoID" == api_name) {
    HandleEnumDateFormats(ip, 2);  // callback is the 2nd parameter
  } else if ("Toolhelp32ReadProcessMemory" == api_name) {
    // call test!ILT+630(_Toolhelp32ReadProcessMemory (0041127b)  [PDB]
    // call dword ptr [test+0x3000 (013c3000)]
    // ds:0023:013c3000={kernel32!Toolhelp32ReadProcessMemory (77601469)}
    HandleToolhelp32ReadProcessMemory(ip, end_offset);
  } else if ("SetUnhandledExceptionFilter" == api_name) {
    // call dword ptr [test!_imp__SetUnhandledExceptionFilter (012d3000)]
    // ds:0023:012d3000={kernel32!SetUnhandledExceptionFilter (775bf5ab)}
    HandleSetUnhandledExceptionFilter(ip);
  } else if ("CreateFileA" == api_name) {
    // jmp     dword ptr [image00400000+0x2000 (00402000)]
    // ds:0023:00402000={kernel32!CreateFileA (7654eb11)}
    HandleCreateFileA(ip);
  } else if ("WriteFileImplementation" == api_name) {
    // call dword ptr [image00400000+0x707c (0040707c)]
    // ds:0023:0040707c={kernel32!WriteFileImplementation (776e54a6)}
    HandleWriteFile(ip);
  } else if ("WriteProcessMemoryStub" == api_name) {
    if (0 == HandleWriteProcessMemory(ip, end_offset)) {
      ProcessSingleInstruction();
    }
  } else if ("OutputDebugStringWStub" == api_name ||
	     "OutputDebugStringAStub" == api_name) {
    // call dword ptr [test!_imp__OutputDebugStringW (00e03004)]
    // ds:0023:00e03004={kernel32!OutputDebugStringWStub (75906f02)}
    HandleOutputDebugString(ip);
  } else if ("EndUpdateResourceA" == api_name) {
    HandleEndUpdateResourceA(ip);
  } else if ("GetLocalTimeStub" == api_name) {
    HandleGetLocalTime(ip);
  } else if ("ZwSetDebugFilterState" == api_name) {
    HandleZwSetDebugFilterState(ip);
  } else if ("NtQueryVirtualMemory" == api_name) {
    HandleNtQueryVirtualMemory(ip);
  } else if ("ZwQueryObject" == api_name){
    HandleZwQueryObject(ip);
  } else {
    // other system APIs
  }

  m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
		     "p",    // step over this API call
		     DEBUG_EXECUTE_NO_REPEAT);  // parent_process.exe

  return DEBUG_STATUS_STEP_OVER;
}

void
EXT_CLASS::HandleAddVectoredExceptionHandler(_In_ ULONG64 ip) {
  // get the handler entry point
  // push    offset test!ILT+300
  //         (?VectoredHandler1YGJPAU_EXCEPTION_POINTERSZ) (00b91131)
  int insn_count = 0;
  ULONG64 last_ip = ip - 1;
  while (true) {
    if (m_addr_code_map.find(last_ip) != m_addr_code_map.end()) {
      ++insn_count;
      if (2 == insn_count) {
	break;
      }
    }
    --last_ip;
  }
  char* insn = m_addr_code_map[last_ip];
  ULONG insn_size = strlen(insn);
  int j = insn_size - 1;
  for (; j > 0; --j) {
    if ('(' == insn[j - 1]) {
      break;
    }
  }
  insn[j + 8] = '\0';
  ULONG64 veh_entry = strtoul(&(insn[j]), 0, 16);
  insn[j + 8] = ')';
  
  // set a software breakpoint on this address
  ULONG bp_id =
    m_BreakpointEx.AddBreakpointOffset(m_Control,
				       veh_entry,
				       DEBUG_BREAKPOINT_CODE,
				       ENUM_BP_2_STR[EXCEPT_HANDLER]);
	
  // make a note
  Int3AddressCode addr_code;
  addr_code.Address = veh_entry;
  addr_code.Id = bp_id;
  m_int3_address_code.push_back(addr_code);

  m_stat_file << "[Anti] " << std::hex << ip
	      << " Set a bp at AddVectoredExceptionHandler() entry:"
	       << veh_entry << std::endl;

  return;
}

void
EXT_CLASS::HandleGetTickCount(_In_ ULONG64 ip) {
  m_need_to_update_tsc = true;
  m_stat_file << "[Anti] " << std::hex << ip
	      << " Handle GetTickCount()" <<  std::endl;
  return;
  /*
    m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
                       "p",
                       DEBUG_EXECUTE_NO_REPEAT);
  */
}

void
EXT_CLASS::HandletimeGetTime(_In_ ULONG64 ip) {
  m_need_to_update_tsc = true;
  m_stat_file << "[Anti] " << std::hex << ip
	      << " Handle timeGetTime()" <<  std::endl;
  return;
}

void
EXT_CLASS::HandleQueryPerformanceCounter(_In_ ULONG64 ip) {
  m_need_to_update_tsc = true;
  // get the LARGE_INTEGER address
  ULONG64 esp_pointer = GetRegisterU64("esp");
  m_QueryPerfCounter_result_address = 0;
  m_Data->ReadVirtualUncached(esp_pointer,
			      &m_QueryPerfCounter_result_address,
			      4,
			      NULL);
  m_stat_file << "[Anti] " << std::hex << ip
	      << " Handle QueryPerformanceCounter()" <<  std::endl;
  
  return;
}

void
EXT_CLASS::HandleCheckRemoteDebuggerPresent(_In_ ULONG64 ip) {
  ULONG64 esp_pointer = GetRegisterU64("esp");
  esp_pointer += 4;
  m_CheckRemoteDebugger_result_addr = 0;
  m_Data->ReadVirtualUncached(esp_pointer,
			      &m_CheckRemoteDebugger_result_addr,
			      4,
			      NULL);
  m_stat_file << "[Anti] " << std::hex << ip
	      << " Handle CheckRemoteDebuggerPresent()" << std::endl;
  return;
}

ULONG
EXT_CLASS::HandleCmp(_In_ ULONG64 ip,
		     _In_ char* instruction,
		     _In_ ULONG second_space_idx,
		     _In_ ULONG third_space_idx,
		     _In_ ULONG disassembled_size) {
  if ('c' == instruction[second_space_idx + 1] &&
      'm' == instruction[second_space_idx + 2] &&
      'p' == instruction[second_space_idx + 3] &&
      ' ' == instruction[second_space_idx + 4]) {
    // cmp
    int j = third_space_idx;
    while (' ' == instruction[++j]);
    if ('b' == instruction[j] &&
	't' == instruction[j + 2] &&
	'p' == instruction[j + 5] &&
	'[' == instruction[j + 9]) {
      // cmp    byte ptr [eax],0CCh    ds:0023:00401042=bb
      // dereference the value at the byte ptr []
      int k = disassembled_size - 1;
      for (; k > j; --k) {
	if (':' == instruction[k - 1]) {
	  break;
	}
      }
      instruction[k + 8] = '\0';
      ULONG64 software_bp_address = strtoul(&(instruction[k]), 0, 16);
      instruction[k + 8] = '=';
      for (auto it = m_int3_address_code.begin();
	   it != m_int3_address_code.end();
	   ++it) {
	if ((*it).Address == software_bp_address) {
	  // disable this breakpoint for a while
	  m_stat_file << "[Anti] " << std::hex << ip
		      << " Client scans for 0xcc byte" << std::endl;
	  m_restore_software_bp_id
	    = m_BreakpointEx.GetBreakpointId(software_bp_address);
	  m_BreakpointEx.DisableBreakpoint(m_restore_software_bp_id);
	  break;
	}
      }
    }
    // set up single stepping
    m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
		       "r tf = 1",    // set up the trap flag
		       DEBUG_EXECUTE_NO_REPEAT);
    return DEBUG_STATUS_GO_HANDLED;
  }
  
  return 0;
}

ULONG
EXT_CLASS::HandleRdtsc(_In_ char* instruction,
		       _In_ ULONG second_space_idx,
		       _In_ ULONG64 ip) {
  if ('r' == instruction[second_space_idx + 1] &&
      'd' == instruction[second_space_idx + 2] &&
      's' == instruction[second_space_idx + 4] &&
      'c' == instruction[second_space_idx + 5]) {
    m_stat_file << "[Anti] " << std::hex << ip
		<< " Update time:rdtsc" << std::endl;
    // rdtsc instruction
    m_need_to_update_tsc = true;
    // set up single stepping
    m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
		       "r tf = 1",    // set up the trap flag
		       DEBUG_EXECUTE_NO_REPEAT);
    return DEBUG_STATUS_GO_HANDLED;
  }

  return 0;
}

ULONG
EXT_CLASS::HandleInt(_In_ ULONG64 ip,
		     _In_ char* instruction,
		     _In_ ULONG second_space_idx,
		     _In_ ULONG64 end_offset) {
  if ('i' == instruction[second_space_idx + 1] &&
      'n' == instruction[second_space_idx + 2] &&
      't' == instruction[second_space_idx + 3] &&
      ' ' == instruction[second_space_idx + 4]) {
    // int X exception, rasied by the client
    // get the interrupt number
    ULONG64 m_int_number = strtoul(&(instruction[second_space_idx + 5]), 0, 16);
    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle int " << m_int_number << std::endl;
    if (3 == m_int_number) {
      m_int3_by_client = true;
      return DEBUG_STATUS_GO_NOT_HANDLED;
    } else if (0x2a == m_int_number) {
      // skip this instruction
      SetRegisterU64("eip", end_offset);  // the next instruction
      m_time_stamp_counter += 2;          // update time stamp counter
      SetRegisterU64("eax", m_time_stamp_counter);
      ProcessSingleInstruction();    // process the next instruction
    } else {
      return DEBUG_STATUS_STEP_OVER;
    }
  }

  return 0;
}

ULONG
EXT_CLASS::HandleMov(_In_ ULONG64 ip,
		     _In_ char* instruction,
		     _In_ ULONG second_space_idx,
		     _In_ ULONG third_space_idx) {
  if ('m' == instruction[second_space_idx + 1] &&
      'o' == instruction[second_space_idx + 2] &&
      'v' == instruction[second_space_idx + 3] &&
      ' ' == instruction[second_space_idx + 4]) {
    // mov     ax,gs
    int j = third_space_idx;
    while (instruction[j] == ' ' ) {
      ++j;
    }
    if ('a' == instruction[j] && 'x' == instruction[j + 1] &&
	'g' == instruction[j + 3] && 's' == instruction[j + 4]) {
      m_stat_file << "[Anti] " << std::hex << ip
		  << " Handle selector GS" << std::endl;
      m_copy_gs = true;
      m_gs = GetRegisterU64("gs");
      return DEBUG_STATUS_STEP_OVER;
    }
  }

  return 0;  // this function does not handle this instruction
}

ULONG
EXT_CLASS::HandleRet(_In_ ULONG64 ip,
		     _In_ char* instruction,
		     _In_ ULONG second_space_idx) {
  if ('r' == instruction[second_space_idx + 1] &&
      'e' == instruction[second_space_idx + 2] &&
      't' == instruction[second_space_idx + 3]) {
    if (m_SEH_resume_address_pointer) {
      // read the resume address
      ULONG64 resume_address = 0;
      m_Data->ReadVirtualUncached(m_SEH_resume_address_pointer,
				  &resume_address,
				  4,
				  NULL);
      // set a software breakpoint on that address
      m_BreakpointEx.AddBreakpointOffset(m_Control,
					 resume_address,
					 DEBUG_BREAKPOINT_CODE,
					 ENUM_BP_2_STR[EXCEPT_RESUME_ADDRESS]);
      m_SEH_resume_address_pointer = 0;

      m_stat_file << "[Anti] " << std::hex << ip
		  << " Set a bp at the EH return address" << std::endl;
      
      return DEBUG_STATUS_STEP_INTO;
    }
  }

  return 0;
}

ULONG
EXT_CLASS::HandleFindWindow(_In_ ULONG64 ip) {
  // esp -> 00401044    ClassName = "ID"
  //        00000000    WindowName = NULL
  // check the matching string, address at [esp]
  ULONG64 esp_pointer = GetRegisterU64("esp");
  ULONG64 stack_value = 0;
  m_Data->ReadVirtualUncached(esp_pointer,
			      &stack_value,
			      4,  // 4 bytes, 32 bit
			      NULL);
  if (stack_value < 0x1000) {
    return 1;    // invalid address
  }
  
  char class_name[17];
  int p = 0;
  for (; p < 17; ++p) {
    BYTE byte;
    m_Data->ReadVirtualUncached(stack_value + p,
				&byte,
				1,  // 1 bytes
				NULL);
    if (byte) {
      class_name[p] = (char)byte;
    } else {
      break;
    }
  }
  
  class_name[p] = '\0';
  std::string class_name_str(class_name);
  if ("OLLYDBG" == class_name_str ||
      "WinDbgFrameClass" == class_name_str ||
      "ID" == class_name_str ||
      "Zeta Debugger" == class_name_str ||
      "Rock Debugger" == class_name_str ||
      "ObsidianGUI" == class_name_str) {

    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle FindWindow():" 
		<< class_name_str << std::endl;

    m_clear_eax = true;
    return 0;
  }
  return 1;
}

ULONG
EXT_CLASS::HandleBlockInput(_In_ ULONG64 ip, _In_ ULONG64 end_offset) {
  m_stat_file << "[Anti] " << std::hex << ip
	      << " Handle BlockInput()" << std::endl;
  std::stringstream command;
  command << "r esp = esp + 4; "
	  << "r eip = 0x" << std::hex << end_offset << "; "
	  << "r eax = 0x" << std::hex << m_BlockInput_counter;
  ++m_BlockInput_counter;
  m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
		     command.str().c_str(),
		     DEBUG_EXECUTE_NO_REPEAT);

  return 0;
}

ULONG
EXT_CLASS::HandleNtQueryInformationProcess(_In_ ULONG64 ip) {
  // get the address that stores the result of ProcessInformation
  // esp -> FFFFFFFF
  //        00000007
  //        0012FF88
  //        00000004
  //        00000000
  ULONG64 esp_pointer = GetRegisterU64("esp");
  ULONG32 process_information_class = 0;
  m_Data->ReadVirtualUncached(esp_pointer + 4,
			      &process_information_class,
			      4,
			      NULL);
  if (0 == process_information_class) {
    // 0: ProcessBasicInformation
    // need to overwrite parent process id with "Explorer.exe"'s.
    m_NtQueryInformationProcess_0_addr = 0;  // base address of type 0
    m_Data->ReadVirtualUncached(esp_pointer + 8,
				&m_NtQueryInformationProcess_0_addr,
				4,
				NULL);
    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle NtQueryInformationProcess(0)" << std::endl;
    return 0;
  }
  
  if ((7 == process_information_class)
      || (0x1e == process_information_class)
      || (0x1f == process_information_class)) {
    // 7: ProcessDebugPort
    // 0x1e: ProcessDebugObjectHandle
    // 0x1f: ProcessDebugFlags
    if (0x1f == process_information_class) {
      m_ProcessDebugFlags = true;
    } else {
      m_ProcessDebugFlags = false;
    }
    m_NtQueryInformationProcess_rlt_addr = 0;
    m_Data->ReadVirtualUncached(esp_pointer + 8,
				&m_NtQueryInformationProcess_rlt_addr,
				4,
				NULL);
    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle NtQueryInformationProcess("
		<< process_information_class
		<< ")" << std::endl;

    return 0;
  }
      
  return 1;
}


ULONG
EXT_CLASS::HandleSetInformationThread(_In_ ULONG64 ip,
				      _In_ ULONG64 end_offset) {
  // ignore this call only when the 4 parameters are valid
  // esp -> FFFFFFFE
  //        00000011
  //        00000000
  //        00000000
  ULONG64 esp_pointer = GetRegisterU64("esp");
  DWORD ThreadHandle = 0;
  DWORD ThreadInformationClass = 0;
  DWORD ThreadInformation = 0;
  DWORD ThreadInformationLength = 0;

  m_Data->ReadVirtualUncached(esp_pointer, &ThreadHandle, 4, NULL);
  m_Data->ReadVirtualUncached(esp_pointer + 4,
			      &ThreadInformationClass,
			      4,
			      NULL);
  m_Data->ReadVirtualUncached(esp_pointer + 8, &ThreadInformation, 4, NULL);
  m_Data->ReadVirtualUncached(esp_pointer + 0xc,
			      &ThreadInformationLength,
			      4,
			      NULL);

  if ((0xfffffffe == ThreadHandle) &&
      (0x11 == ThreadInformationClass) &&
      (0 == ThreadInformation) &&
      (0 == ThreadInformationLength)) {
    // skip this call
    std::stringstream command;
    command << "r esp = esp + 0x10; "    // clean up stack
	    << "r eip = 0x" << std::hex << end_offset << "; "  // the next insn
	    << "r eax = 0;";    // STATUS_SUCCESS
    m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
		       command.str().c_str(),
		       DEBUG_EXECUTE_NO_REPEAT);

    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle SetInformationThread()" << std::endl;
    
    return 0;
  }

  return 1;
}

ULONG
EXT_CLASS::HandleSwitchDesktop(_In_ ULONG64 ip, _In_ ULONG64 end_offset) {
  // ignore this call if the parameter in stack is non-zero
  // esp -> 0000003C
  ULONG64 esp_pointer = GetRegisterU64("esp");
  DWORD hDesktop = 0;
  m_Data->ReadVirtualUncached(esp_pointer, &hDesktop, 4, NULL);
  if (hDesktop) {
    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle SwitchDesktop()" << std::endl;
    
    std::stringstream command;
    command << "r esp = esp + 4; "    // clean up stack
	    << "r eip = 0x" << std::hex << end_offset << "; "  // the next insn
	    << "r eax = 1;";          // the function succeeds
    m_Control->Execute(DEBUG_OUTCTL_THIS_CLIENT,
		       command.str().c_str(),
		       DEBUG_EXECUTE_NO_REPEAT);
    
    return 0;
  }

  return 1;
}

ULONG
EXT_CLASS::HandleEnumDateFormats(_In_ ULONG64 ip, int call_back_index) {
  // set a breakpoint on the callback address if it is in user code
  // esp -> 00401061
  ULONG64 esp_pointer = GetRegisterU64("esp");
  ULONG32 callback_address = 0;
  m_Data->ReadVirtualUncached(esp_pointer + call_back_index * 4,
			      &callback_address,
			      4,
			      NULL);
  if (0x1000 < callback_address && callback_address < 0x4fffffff) {
    // TODO: use image base as the lower bound
    ULONG bp_id
      = m_BreakpointEx.AddBreakpointOffset(m_Control,
					   callback_address,
					   DEBUG_BREAKPOINT_CODE,
					   ENUM_BP_2_STR[ENUM_CALLBACK]);
    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle Enum()" << std::endl;
    
    // make a note
    Int3AddressCode addr_code;
    addr_code.Address = callback_address;
    addr_code.Id = bp_id;
    m_int3_address_code.push_back(addr_code);

    return 0;
  }

  return 1;
}

ULONG
EXT_CLASS::HandleToolhelp32ReadProcessMemory(_In_ ULONG64 ip,
					     _In_ ULONG64 end_offset) {
  // store the original byte after this instruction
  // esp -> 000014B8    ProcessID = 14B8
  //        00401057    BaseAddress to read
  //        0012FF88    Buffer to store
  //        00000001    Size = 1 byte
  //        00000000    BytesRead = NULL
  ULONG64 esp_pointer = GetRegisterU64("esp");
  ULONG32 BaseAddress = 0;
  m_Data->ReadVirtualUncached(esp_pointer + 4, &BaseAddress, 4, NULL);
  if (end_offset == BaseAddress) {
    m_Data->ReadVirtualUncached(esp_pointer + 8,
				&m_Toolhelp32_result_addr,
				4,
				NULL);
    
    m_Data->ReadVirtualUncached(BaseAddress,
				&m_Toolhelp32_original_byte,
				1,
				NULL);
    if (0xcc == m_Toolhelp32_original_byte) {
      m_stat_file << "[Todo] " << std::hex << ip
		  << " The original byte read from Toolhelp32ReadProcessMemory "
		  << "is 0xcc. Error. Exit." << std::endl;
      Exit(1);
    }

    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle Toolhelp32ReadProcessMemory()" << std::endl;
    return 0;
  }

  return 1;  // this call does not read the byte after itself
}


ULONG
EXT_CLASS::HandleSetUnhandledExceptionFilter(_In_ ULONG64 ip) {
  // esp -> 0040105E    Filter
  ULONG64 esp_pointer = GetRegisterU64("esp");
  ULONG32 Filter = 0;
  m_Data->ReadVirtualUncached(esp_pointer, &Filter, 4, NULL);
  if ((0x1000 < Filter ) && (Filter) < 0x4fffffff) {
    m_UnhandledExceptFilterAddress = Filter;

    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle UnhandledExceptionFilter()" << std::endl;
    
    return 0;
  }

  return 1;
}

ULONG
EXT_CLASS::HandleRtlProcessFlsData(_In_ ULONG64 ip) {
  // set a breakpoint at the flow destination
  m_stat_file << "[Anti] " << std::hex << ip
	      << " Handle RtlProcessFlsData()" << std::endl;
  return 0;
}

ULONG
EXT_CLASS::HandleCreateFileA(_In_ ULONG64 ip) {
  // 00403000  FileName = "C:\...\unit_tests\api_CreateFile.exe"
  // 80000000  DesiredAccess = GENERIC_READ
  // 00000000  ShareMode = 0
  // 00000000  pSecurity = NULL
  // 00000003  CreationDistribution = OPEN_EXISTING
  // 00000000  Attributes = 0
  // 00000000  hTemplate = NULL
  ULONG64 esp_pointer = GetRegisterU64("esp");
  ULONG32 desired_access = 0;
  ULONG32 share_mode = 0;
  ULONG32 creation_distribution = 0;
  m_Data->ReadVirtualUncached(esp_pointer + 4, &desired_access, 4, NULL);
  m_Data->ReadVirtualUncached(esp_pointer + 8, &share_mode, 4, NULL);
  m_Data->ReadVirtualUncached(esp_pointer + 0x10,
			      &creation_distribution,
			      4,
			      NULL);
  if (desired_access == GENERIC_READ &&
      share_mode == 0 &&
      creation_distribution == OPEN_EXISTING) {
    m_make_CreateFile_succeeded =  true;
    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle CreateFileA()" << std::endl;
  }
  return 0;
}

ULONG
EXT_CLASS::HandleWriteFile(_In_ ULONG64 ip) {
  // record the content written
  // esp -> 00000007    hFile = 00000007
  //  +4    0012FACC    Buffer = 0012FACC -> 75 ('u')
  //  +8    0000003e    Size = 62
  //  +0xc  0012FED4    pBytesWritten = 0012FED4 -> 0
  //        00000000    pOverlapped = NULL
  ULONG64 esp_pointer = GetRegisterU64("esp");
  ULONG64 lpBuffer = 0;
  ULONG32 nNumberOfBytesToWrite = 0;
  m_Data->ReadVirtualUncached(esp_pointer + 4, &lpBuffer, 4, NULL);
  m_Data->ReadVirtualUncached(esp_pointer + 8, &nNumberOfBytesToWrite, 4, NULL);

  m_stat_file << "[Info] " << std::hex << ip
	      << " Log WriteFile()" << std::endl;

  for (int i = 0; i < (int)nNumberOfBytesToWrite; ++i) {
    BYTE original_byte = 0;
    m_Data->ReadVirtualUncached(lpBuffer + i,
				&original_byte,
				1,
				NULL);
  }

  return 0;
}

ULONG
EXT_CLASS::HandleWriteProcessMemory(_In_ ULONG64 ip, _In_ ULONG64 end_offset) {
  // we write the memory instead and skip the call
  // 0012FF78  FFFFFFFF  |hProcess = INVALID_HANDLE_VALUE
  // 0012FF7C  00401025  |BaseAddress = api_WriteProcessMemory.401025
  // 0012FF80  00401026  |Buffer = 00401026
  // 0012FF84  00000001  |Size = 1
  // 0012FF88  00000000  |pBytesWritten = NULL
  ULONG64 esp_pointer = GetRegisterU64("esp");
  ULONG32 handle = 0;
  ULONG32 base = 0;
  ULONG32 buffer = 0;
  ULONG32 size = 0;
  m_Data->ReadVirtual(esp_pointer, &handle, 4, NULL);
  m_Data->ReadVirtual(esp_pointer + 4, &base, 4, NULL);
  m_Data->ReadVirtual(esp_pointer + 8, &buffer, 4, NULL);
  m_Data->ReadVirtual(esp_pointer + 0xc, &size, 4, NULL);

  if (-1 == handle && end_offset == base) {
    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle WriteProcessMemory()" << std::endl;
    
    for (ULONG32 i = 0; i < size; ++i) {
      // may be overlapped
      BYTE b = 0;
      m_Data->ReadVirtual(buffer + i, &b, 1, NULL);
      m_Data->WriteVirtual(base + i, &b, 1, NULL);
    }
    SetRegisterU64("eax", 1);            // nonzero: successful
    SetRegisterU64("eip", end_offset);   // skip this instruction
    SetRegisterU64("esp", esp_pointer + 0x14);    // clean stack

    return 0;
  }

  return 1;
}

ULONG
EXT_CLASS::HandleOutputDebugString(_In_ ULONG64 ip) {
  m_stat_file << "[Anti] " << std::hex << ip
	      << " Handle OutputDebugString()" << std::endl;
  m_need_to_update_last_error = true;

  return 0;
}

ULONG
EXT_CLASS::HandleEndUpdateResourceA(_In_ ULONG64 ip) {
  m_stat_file << "[Anti] " << std::hex << ip
	      << " Handle EndUpateResourceA()" << std::endl;
  m_make_EndUpdateResourceA_succeeded = true;  // unsafe?
  
  return 0;
}

ULONG
EXT_CLASS::HandleGetLocalTime(_In_ ULONG64 ip) {
  // get the address of return value (SystemTime)
  m_update_GetLocalTime = true;
  ULONG64 esp_pointer = GetRegisterU64("esp");
  m_Data->ReadVirtualUncached(esp_pointer, &m_lpSystemTime, 4, NULL);
  
  m_stat_file << "[Anti] " << std::hex << ip
	      << " Handle GetLocalTime()" << std::endl;
    
  return 0;
}

ULONG
EXT_CLASS::HandleZwSetDebugFilterState(_In_ ULONG64 ip) {
  m_randomize_eax = true;
  m_stat_file << "[Anti] " << std::hex << ip
	      << " Handle ZwSetDebugFilterState()" << std::endl;
  
  return 0;
}

ULONG
EXT_CLASS::HandleNtQueryVirtualMemory(_In_ ULONG64 ip) {
  // get the output buffer pointer from the stack
  // FFFFFFFF  _In_       HANDLE                   ProcessHandle
  // 00000000  _In_opt_   PVOID                    BaseAddress
  // 00000001  _In_       MEMORY_INFORMATION_CLASS MemoryInformationClass
  // 001C0000  _Out_      PVOID                    MemoryInformation
  // 00001000  _In_       SIZE_T                   MemoryInformationLength
  // 00000000  _Out_opt_  PSIZE_T                  ReturnLength
  ULONG64 esp_pointer = GetRegisterU64("esp");
  ULONG32 process_handle = 0;
  ULONG32 mem_info_class = 0;
  m_Data->ReadVirtualUncached(esp_pointer, &process_handle, 4, NULL);
  m_Data->ReadVirtualUncached(esp_pointer + 8, &mem_info_class, 4, NULL);
  m_Data->ReadVirtualUncached(esp_pointer + 0xc, &m_mem_info, 4, NULL);
  m_Data->ReadVirtualUncached(esp_pointer + 0x10, &m_mem_info_length, 4, NULL);

  if (0xffffffff == process_handle
      && 1 == mem_info_class
      && m_mem_info > 0
      && m_mem_info_length > 0) {
    m_handle_NtQueryVirtualMemory = true;
    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle NtQueryVirtualMemory()" << std::endl;
      
    return 0;
  }

  return 1;    // not applicable
}

ULONG
EXT_CLASS::HandleZwQueryObject(_In_ ULONG64 ip) {
  // get the output buffer pointer from the stack
  // 00000000  IN  HANDLE                   ObjectHandle
  // 00000003  IN  OBJECT_INFORMATION_CLASS ObjectInformationClass
  // 002C0000  OUT PVOID                    ObjectInformation
  // 00001000  IN  ULONG                    Length
  // 00000000  OUT PULONG                   ResultLength
  ULONG64 esp_pointer = GetRegisterU64("esp");
  ULONG32 object_information_class;
  m_Data->ReadVirtual(esp_pointer + 4, &object_information_class, 4, NULL);
  m_Data->ReadVirtual(esp_pointer + 8, &m_object_information_buffer, 4, NULL);
  
  if(3 == object_information_class && 0 != m_object_information_buffer) {
    // 3: ObjectAllInformation
    m_check_eax_for_ZwQueryObject = true;
    m_stat_file << "[Anti] " << std::hex << ip
		<< " Handle ZwQueryObject()" << std::endl;
    return 0;
  }

  return 1;
}

void
EXT_CLASS::PostHandleZwQueryObject() {
  // find and overwrite "DebugObject" string in the output buffer
  // get the number of objects
  ULONG32 num_objects = 0;
  ULONG32 i = 0;
  m_Data->ReadVirtual(m_object_information_buffer, &num_objects, 4, NULL);
  m_object_information_buffer += 4;    // lodsd dowrd ptr[esi]
  for (ULONG32 i = 1; i <= num_objects; ++i) {
    ULONG32 length = 0;
    m_Data->ReadVirtual(m_object_information_buffer, &length, 4, NULL);
    m_object_information_buffer += 4;
    length = length & 0x0000ffff;    // movzx edx, ax
    ULONG32 name_address = 0;
    m_Data->ReadVirtual(m_object_information_buffer, &name_address, 4, NULL);
    m_object_information_buffer = name_address;
    if (0x16 == length) {
      char debug[] = {'D', 'e', 'b', 'u', 'g', 'O', 'b', 'j', 'e', 'c', 't'};
      BOOLEAN match = true;
      for (int j = 0; j < 11; ++j) {
	BYTE letter;
	m_Data->ReadVirtual(m_object_information_buffer + 2 * j,
			    &letter,
			    1,
			    NULL);
	if (debug[j] != letter) {
	  match = false;
	  break;
	}
      }
      if (match) {
	// overwrite "DebugObject" string
	BYTE new_value = 0x55;
	m_Data->WriteVirtualUncached(m_object_information_buffer,
				     &new_value,
				     1,
				     NULL);
	break;
      }
    } else {
      m_object_information_buffer += (length + 4);
      m_object_information_buffer = m_object_information_buffer & 0xfffffffc;
    }
  }
}
