/* Copyright (C) Hao Shi 2014 */

#include "event_callbacks.h"
#include <stdio.h>

STDMETHODIMP_(ULONG)
EventCallbacks::AddRef(
  THIS
  ) {
  // This class is designed to be static so
  // there's no true refcount.
  return 1;
}

STDMETHODIMP_(ULONG)
EventCallbacks::Release(
  THIS
  ) {
  // This class is designed to be static so
  // there's no true refcount.
  return 0;
}

STDMETHODIMP
EventCallbacks::GetInterestMask(
  THIS_
  _Out_ PULONG mask
  ) {
  // we are interested in the following events
  *mask =
    DEBUG_EVENT_BREAKPOINT     |
    DEBUG_EVENT_EXCEPTION      |
    DEBUG_EVENT_CREATE_PROCESS |
    DEBUG_EVENT_CREATE_THREAD  |
    DEBUG_EVENT_LOAD_MODULE    |
    DEBUG_EVENT_SESSION_STATUS;
  return S_OK;
}

STDMETHODIMP
EventCallbacks::Breakpoint(THIS_ _In_ PDEBUG_BREAKPOINT breakpoint) {
  // callback will receive the event first
  // if it returns DEBUG_STATUS_BREAK, WaitForEvent() will break and we
  // need to check the last event there.
  return DEBUG_STATUS_BREAK;
}

STDMETHODIMP
EventCallbacks::Exception(
  THIS_
  _In_ PEXCEPTION_RECORD64 exception,
  _In_ ULONG first_chance
  ) {
  // We want to handle these exceptions on the first
  // chance to make it look like no exception ever
  // happened.  Handling them on the second chance would
  // allow an exception handler somewhere in the app
  // to be hit on the first chance.

  return DEBUG_STATUS_BREAK;
}

STDMETHODIMP
EventCallbacks::CreateProcess(
  THIS_
  _In_ ULONG64 image_file_handle,
  _In_ ULONG64 handle,
  _In_ ULONG64 base_offset,
  _In_ ULONG module_size,
  _In_ PCSTR module_name,
  _In_ PCSTR image_name,
  _In_ ULONG check_sum,
  _In_ ULONG time_data_stamp,
  _In_ ULONG64 initial_thread_handle,
  _In_ ULONG64 thread_data_offset,
  _In_ ULONG64 start_offset) {
  UNREFERENCED_PARAMETER(image_file_handle);
  UNREFERENCED_PARAMETER(handle);
  UNREFERENCED_PARAMETER(module_size);
  UNREFERENCED_PARAMETER(module_name);
  UNREFERENCED_PARAMETER(check_sum);
  UNREFERENCED_PARAMETER(time_data_stamp);
  UNREFERENCED_PARAMETER(initial_thread_handle);
  UNREFERENCED_PARAMETER(thread_data_offset);
  UNREFERENCED_PARAMETER(start_offset);
    
  // The process is now available for manipulation.
  (*m_stat_file) << "[Info] New process created. "
		 << "Handle:" << std::hex << handle << ". "
		 << "Module:" << module_name << ". "
		 << "Initial_thread_handle:"
		 << std::hex << initial_thread_handle << ". "
		 << "Start_offset:" << std::hex << start_offset
		 << std::endl;

  return DEBUG_STATUS_BREAK;
}

STDMETHODIMP
EventCallbacks::CreateThread(
    THIS_
    _In_ ULONG64 Handle,
    _In_ ULONG64 DataOffset,
    _In_ ULONG64 StartOffset) {
  UNREFERENCED_PARAMETER(Handle);
  UNREFERENCED_PARAMETER(DataOffset);
  UNREFERENCED_PARAMETER(StartOffset);
  (*m_stat_file) << "[Info] New thread created. "
		 << "Handle:" << std::hex << Handle << ". "
		 << "StartOffset: " << StartOffset << std::endl;

  return DEBUG_STATUS_BREAK;
}

STDMETHODIMP
EventCallbacks::LoadModule(
  THIS_
  _In_ ULONG64 image_file_handle,
  _In_ ULONG64 base_offset,
  _In_ ULONG module_size,
  _In_ PCSTR module_name,
  _In_ PCSTR image_name,
  _In_ ULONG check_sum,
  _In_ ULONG time_date_stamp
  ) {
  UNREFERENCED_PARAMETER(image_file_handle);
  UNREFERENCED_PARAMETER(module_size);
  UNREFERENCED_PARAMETER(module_name);
  UNREFERENCED_PARAMETER(check_sum);
  UNREFERENCED_PARAMETER(time_date_stamp);

  (*m_stat_file) << "[Info] Moduled loaded:" << module_name << std::endl;
  
  return DEBUG_STATUS_BREAK;
}


STDMETHODIMP
EventCallbacks::SessionStatus(
  THIS_
  _In_ ULONG session_status
  ) {
  // A session isn't fully active until WaitForEvent
  // has been called and has processed the initial
  // debug events.  We need to wait for activation
  // before we query information about the session
  // as not all information is available until the
  // session is fully active.  We could put these
  // queries into CreateProcess as that happens
  // early and when the session is fully active, but
  // for example purposes we'll wait for an
  // active SessionStatus callback.
  // In non-callback applications this work can just
  // be done after the first successful WaitForEvent.
  if (session_status != DEBUG_SESSION_ACTIVE) {
    return S_OK;
  }

  return DEBUG_STATUS_BREAK;
}

